/**
 * ds2caxml - extract formats into named xml file.
 *
 * arguments:
 *  catalog   - the catalog from which to extact formats
 *  fileref   - the file to contain the format xml
 *  encoding  - the encoding (e.g.: utf8) for the generated file
 *  formats   - the names of formats to select (empty means all)
 *  useview   - yes means use a view; no means use a temp file
 *
 * example usage:
 *  %ds2caxml(work.formats,myxml2,utf8,formats);
 */

******************************************************;
%macro ds2caxml(catalog,fileref,encoding,formats,useview=no);
%local dsname viewname varlenchkoption;
%if &useview=no %then %do;
proc format library=&catalog cntlout=temp;
%if %length(&formats) ne 0 %then %do;
  select &formats;
%end;
; run;
%let dsname=temp;
%let viewname=_null_;
%end;
%else %do;
%let dsname=outview;
%let viewname=new/view=outview;
%end;

%let varlenchkoption=%sysfunc(getoption(varlenchk));
options varlenchk=nowarn;
filename mylog temp;
proc printto log=mylog new; run;

data &viewname;
     length fmtname $32 start end label $512 min max default length fuzz 8
            prefix $1 mult 8 fill $1 noedit 8 type sexcl eexcl $1 hlo $8 decsep dig3sep datatype language $1;
     set &dsname end=eof;
     by fmtname type notsorted;
     file &fileref;
     retain nobs 2;
     length line $32767;
     if _n_=1 then do;
        array varname[21] $8 _temporary_;
        array varlabel[21] $200 _temporary_;
        array varlen[21] _temporary_;
        array vartype[21] $1 _temporary_;
        length names $8;
        i=0;
        do names='fmtname','start','end','label','min','max','default','length','fuzz','prefix','mult',
                 'fill','noedit','type','sexcl','eexcl','hlo','decsep','dig3sep','datatype','language';
           i+1;
           varname[i]=vnamex(names);
           varlabel[i]=vlabelx(names);
           varlen[i]=vlengthx(names);
           vartype[i]=vtypex(names);
           end;
        offset = 0;
        do i=1 to dim(varlen); offset + varlen[i]; end;
        link write_header;
        offset = 0;
        do i=1 to dim(varname);
           link write_desc;
           end;
        session_encoding = getoption('ENCODING');
        retain session_encoding;
        end;

     line=' ';      link putline;
     line='<ROW>';  link putline;

     if first.fmtname then do;
        link complete_obs;
        end;
     /* YESNO. and $YESNO. have the same FMTNAME but different TYPE. */
     else if first.type then do;
        link complete_obs;
        end;
     /* YESNO. and $YESNO. have the same FMTNAME but different TYPE. */
     else if first.type then do;
        link complete_obs;
        if not eof then delete;
        end;
     else do;
        link reduced_obs;
        end;
     line='</ROW>'; link putline;
     if not eof then delete;

     if eof;
        line='</TABLE>'; link putline;
        line='</LIBRARY>'; link putline;
        delete;

     return;

write_desc:;
     line=' '; link putline;
     line=cats('<COLUMN name="',upcase(varname[i]),'" order="',i,'" label="',varlabel[i],'">'); link putline;
     line=cats('<TYPE>',ifc(vartype[i]='C','character','numeric'),'</TYPE>'); link putline;
     line=cats('<DATATYPE>',ifc(vartype[i]='C','string','float'),'</DATATYPE>');link putline;
     if vartype[i]='C' then do;
        line=cats('<LENGTH>',varlen[i],'</LENGTH>');link putline;
        end;
     line=cats('<Offset>',offset,'</Offset>');link putline;
     line='</COLUMN>'; link putline;
     offset+varlen[i];
     return;

write_header:;
     crdttm = put(datetime(),e8601dt19.);
                   line='<?xml version="1.0" encoding="windows-1252" ?>';
%if %length(&encoding) ne 0 %then %do;
     line = tranwrd(line,'windows-1252',"&encoding.");
%end;
     link putline; line='<?xml-stylesheet type="text/xsl" href="SUVformats.xsl"?>';
     link putline; line='<LIBRARY type="EXPORT" version="SUV">';
     link putline; line='<HEADER>';
     link putline; line='<Provider>SAS Institute Inc.</Provider>';
     link putline; line="<Version>&sysver.</Version>";
     link putline; line="<VersionLong>&sysvlong4.</VersionLong>";
     link putline; line=cats('<CreationDateTime>',crdttm,'</CreationDateTime>');
     link putline; line='</HEADER>';
     link putline; line=' ';
     link putline; line='<TABLE name="cntlout">';
     link putline; line='<TABLE-HEADER>';
     link putline; line='<Provider>SAS Institute Inc.</Provider>';
     link putline; line="<Version>&sysver.</Version>";
     link putline; line="<VersionLong>&sysvlong4.</VersionLong>";
     link putline; line=cats('<CreationDateTime>',crdttm,'</CreationDateTime>');
     link putline; line=cats('<ModifiedDateTime>',crdttm,'</ModifiedDateTime>');
     link putline; line=' ';
     link putline; line='<Protection />';
     link putline; line='<DataSetType />';
     link putline; line='<DataRepresentation />';
     link putline; line='<Encoding>windows-1252</Encoding>';
%if %length(&encoding) ne 0 %then %do;
     line = tranwrd(line,'windows-1252',"&encoding.");
%end;
     link putline; line='<ReleaseCreated />';
     link putline; line='<HostCreated />';
     link putline; line=cats('<Filename>',pathname('myxml2'),'</Filename>');
     link putline; line=' ';
     link putline; line=cats('<Observations length="',offset,'" />');
     link putline; line=cats('<Compression compressed="No" number="1" length="',offset,'" />');
     link putline; line='<Variables number="21" />';
     link putline; line='</TABLE-HEADER>';
     link putline;
     return;

complete_obs:;
     line = FMTNAME; link xmlencode;
     line = cats('<FMTNAME>',line,'</FMTNAME>');
     link putline;
     line = START; link xmlencode;
     line = cats('<START>',line,'</START>');
     link putline;
     line = END; link xmlencode;
     line = cats('<END>',line,'</END>');
     link putline;
     line = LABEL; link xmlencode;
     line = cats('<LABEL>',line,'</LABEL>');
     link putline;
     line = MIN; link xmlencode;
     line = cats('<MIN>',line,'</MIN>');
     link putline;
     line = MAX; link xmlencode;
     line = cats('<MAX>',line,'</MAX>');
     link putline;
     line = DEFAULT; link xmlencode;
     line = cats('<DEFAULT>',line,'</DEFAULT>');
     link putline;
     line = LENGTH; link xmlencode;
     line = cats('<LENGTH>',line,'</LENGTH>');
     link putline;
     line = FUZZ; link xmlencode;
     line = cats('<FUZZ>',line,'</FUZZ>');
     link putline;
     line = PREFIX; link xmlencode;
     line = cats('<PREFIX>',line,'</PREFIX>');
     link putline;
     line = MULT; link xmlencode;
     line = cats('<MULT>',line,'</MULT>');
     link putline;
     line = FILL; link xmlencode;
     line = cats('<FILL>',line,'</FILL>');
     link putline;
     line = NOEDIT; link xmlencode;
     line = cats('<NOEDIT>',line,'</NOEDIT>');
     link putline;
     line = TYPE; link xmlencode;
     line = cats('<TYPE>',line,'</TYPE>');
     link putline;
     line = SEXCL; link xmlencode;
     line = cats('<SEXCL>',line,'</SEXCL>');
     link putline;
     line = EEXCL; link xmlencode;
     line = cats('<EEXCL>',line,'</EEXCL>');
     link putline;
     line = HLO; link xmlencode;
     line = cats('<HLO>',line,'</HLO>');
     link putline;
     line = DECSEP; link xmlencode;
     line = cats('<DECSEP>',line,'</DECSEP>');
     link putline;
     line = DIG3SEP; link xmlencode;
     line = cats('<DIG3SEP>',line,'</DIG3SEP>');
     link putline;
     line = DATATYPE; link xmlencode;
     line = cats('<DATATYPE>',line,'</DATATYPE>');
     link putline;
     line = LANGUAGE; link xmlencode;
     line = cats('<LANGUAGE>',line,'</LANGUAGE>');
     link putline;

     return;

reduced_obs:;
     line = START; link xmlencode;
     line = cats('<START>',line,'</START>');
     link putline;

     if start ne end then do;
        line = END; link xmlencode;
        line = cats('<END>',line,'</END>');
        link putline;
        end;

     line = LABEL; link xmlencode;
     line = cats('<LABEL>',line,'</LABEL>');
     link putline;

     if hlo ne ' ' then do;
        line = HLO; link xmlencode;
        line = cats('<HLO>',line,'</HLO>');
        link putline;
        end;
     return;

xmlencode:;
     /* encode xml entities */
     line = tranwrd(line, '&', '&amp;');
     line = tranwrd(line, '>', '&gt;');
     line = tranwrd(line, '<', '&lt;');
     line = tranwrd(line, '"', '&quot;');
     line = tranwrd(line, '''', '&apos;');
     return;

putline:;
     if line=' ' then return;
     j=index(line,'><');
*    if j then line=substr(line,1,j-1)||' missing=" " />';
     if j then line=substr(line,1,j-1)||'/>';
%if %length(&encoding) ne 0 %then %do;
     line=kcvt(line,session_encoding,"&encoding.");
%end;
     put line;
     return;
     run;

%if &useview=yes %then %do;
proc format library=&catalog cntlout=outview;
%if %length(&formats) ne 0 %then %do;
  select &formats;
%end;
; run;
%end;
%else %do;
proc delete data=temp; run;
%end;

proc printto; run;
data _null_; infile mylog;
     input;
     if _infile_=:'WARNING' then do;
        input;
        return;
        end;
     put _infile_;
     run;
filename mylog clear;

options varlenchk=&varlenchkoption;

%mend ds2caxml;
